(unit/sig context^
  (import)
  
  ; context = (make-context nat nat size color)
  
  ;(define-struct context (code underline size color))
  ; use define-values instead because my sensible naming convetion was rejected  
  
  (define-values (struct:context real-make-context context?
                                 context-code set-context-code!
                                 get-underline set-context-underline!
                                 get-size set-context-size!
                                 get-color set-context-color!)
    (struct context (code underline size color)))

  ; initial-context : context
  (define initial-context (real-make-context 0 0 'root 'root))

  ; meaning-tree = meaning | (cons sym (listof meaning-tree))
  ; meaning = (make-meaning context str)
  (define-struct meaning (context str))
  (define-struct (meaning/space struct:meaning) ())
  

  ; dave tucker wanted this
  (define initial-meaning (make-meaning initial-context ""))
  
  (define b-flag  #b00001)
  (define em-flag #b00010)
  (define i-flag  #b00100)
  (define s-flag  #b01000)
  (define tt-flag #b10000)
  
  (define not-b-flag (bitwise-not b-flag))
  (define not-em-flag (bitwise-not em-flag))
  (define not-i-flag (bitwise-not i-flag))
  (define not-s-flag (bitwise-not s-flag))
  (define not-tt-flag (bitwise-not tt-flag))
  
  ; make-is-x? : nat -> context -> bool
  (define (make-is-x? flag)
    (lambda (x)
      (= flag (bitwise-and flag (context-code x)))))
  
  ; context -> bool
  (define is-b? (make-is-x? b-flag))
  (define is-em? (make-is-x? em-flag))
  (define is-i? (make-is-x? i-flag))
  (define is-s? (make-is-x? s-flag))
  (define is-tt? (make-is-x? tt-flag))
  
  
  ; curry-ior : nat -> nat -> nat
  (define (curry-ior x)
    (lambda (y) (bitwise-ior x y)))
  
  ; curry-xor : nat -> nat -> nat
  (define (curry-xor x)
    (lambda (y) (bitwise-xor x y)))
  
  ; context-open-tag : context sym -> context
  (define (context-open-tag x tag)
    (case tag
      [(B) (update-code (curry-ior b-flag) x)]
      [(EM) (if (is-s? x)
                x
                (update-code (curry-xor em-flag) x))]
      [(I) (update-code (curry-ior i-flag) x)]
      [(PL) (real-make-context 0 0 (get-size x) (get-color x))]
      [(S) (update-code (lambda (code)
                          (bitwise-and not-em-flag (bitwise-ior s-flag code)))
                        x)]
      [(TT) (update-code (curry-ior tt-flag) x)]
      [(U) (let ([underline (get-underline x)])
             (if (< underline 3)
                 (real-make-context (context-code x) (add1 underline)
                                    (get-size x) (get-color x))
                 x))]
      [(r g b c m y k w)
       (real-make-context (context-code x) (get-underline x) (get-size x) tag)]
      [(|0| |1| |2| |3| |4| |5| |6| |7| |8| |9|)
       (real-make-context (context-code x) (get-underline x) tag (get-color x))]
      [else (error 'context-open-tag "Unknown tag ~a" tag)]))
  
  ; update-code : (nat -> nat) context -> context
  (define (update-code op x)
    (real-make-context
     (op (context-code x))
     (get-underline x) (get-size x) (get-color x)))
  
  ; space-context : context -> context
  (define (space-context x)
    (let ([u (get-underline x)])
      (real-make-context (bitwise-and tt-flag (context-code x))
                         u
                         (get-size x)
                         (if (zero? u)
                             'w
                             (get-color x)))))
  
  ; display-context : context -> void
  (define (display-context x)
    (printf "#(struct:context B=~a EM=~a I=~a S=~a TT=~a U=~a color=~a size=~a)"
            (is-b? x) (is-em? x) (is-i? x) (is-s? x) (is-tt? x)
            (get-underline x) (get-color x) (get-size x)))
  
  
  ; context-distances : (vectorof meaning) -> (vector nat)
  ; return vector is one element shorter than the input vector
  (define (context-distances v)
    (let ([result (make-vector (sub1 (vector-length v)))])
      (let loop ([i (sub1 (vector-length result))]
                 [total-dist 0])
        (when (>= i 0)
          (let ([new-total (+ total-dist
                              (meaning-dist (vector-ref v (add1 i))
                                            (vector-ref v i)))])
            (vector-set! result i new-total)
            (loop (sub1 i) new-total))))
      result))
  
  ; meaning-dist : meaning meaning -> nat
  (define (meaning-dist x y)
    (let* ([x (meaning-context x)]
           [y (meaning-context y)]
           [code-diff (bitwise-xor (context-code x)
                                   (context-code y))])
      (+ (if (is-b? code-diff) 1 0)
         (if (is-tt? code-diff) 1 0)
         (if (is-em? code-diff) 1 0)
         (if (is-s? code-diff) 1 0)
         (if (is-i? code-diff) 1 0)
         (abs (- (get-underline x) (get-underline y)))
         (if (eq? (get-size x) (get-size y)) 0 1)
         (if (eq? (get-color x) (get-color y)) 0 1))))
  )
