(load-relative "tags.ss")


(define (full-simple-search! meaning-vec tag-selection-function answer-space)
  (full-simple-search-help meaning-vec 0 initial-stack tag-selection-function
                           answer-space))

(define (full-simple-search-help meaning-vec pos stack t-s-f answer-space)
  (cond
    ((= pos (vector-length meaning-vec)) (set-answer! answer-space
                                                      pos
                                                      (last-move stack)
                                                      '()))
    (else (let ((step (one-step! meaning-vec pos stack t-s-f
                                 answer-space)))
            ;(display "part of answer is:")
            ;(display (car step))
            ;(display (meaning-str (vector-ref meaning-vec pos)))
            ;(newline)
            
            (full-simple-search-help meaning-vec
                                     (+ 1 pos)
                                     (cdr step)
                                     t-s-f
                                     answer-space)))))


;; This is a function the does one step of the algorithm
;; it relies on begin passed a tag-selection function
;; which chooses somehow one tag list out of the 
;; tag-possibilities given to us by delta.
; one-step: returns a (tag-list . Stack) <--- a pair!, use car/cdr
;; a tag-list will now have it's first element be a count of closings.
(define (one-step! meaning-vec pos stack tag-selection-function
                   answer-space)
  (let* ((choices (delta stack (pull-context meaning-vec pos) (vector-ref meaning-vec pos)))
         (selected (tag-selection-function choices)))
    (set-answer! answer-space pos (caar selected) (cdar selected))
    (cons
     (car selected)
     (cdr selected))))
     
     
     

;;delta: stack X new-context -> list of (tag list . stack)
(define (delta stack new-context new-meaning)
  (delta-help stack
              new-context
              new-meaning
              0))

(define (delta-help stack goal-context new-meaning close-count)
  (delta-help/k (lambda (x) x) stack goal-context new-meaning close-count))

(define (delta-help/k k stack goal-context new-meaning close-count)
  (let ((top-context (first stack)))
    (let ((new-tags (tags-to-diff-context
                     top-context goal-context (meaning/space? new-meaning))))
      (cond
       ((empty? (rest stack))
	(k 
	 (list-tag-stack-pairs new-tags close-count
			       stack)))
       (else
	(delta-help/k
	 (lambda (a)
	   (k
	    (append (list-tag-stack-pairs new-tags close-count
					  stack)
		    a)))
	 (pop-move stack)
	 goal-context
	 new-meaning
	 (+ 1 close-count)))))))

(define (list-tag-stack-pairs open-tags close-count stack)
  (list-tag-stack-pairs-help
   (map (lambda (x) (cons close-count x)) open-tags)
   stack))

(define (list-tag-stack-pairs-help full-tags stack)
  (cond
    ((empty? full-tags) '())
    (else (cons (cons (first full-tags)
                      (several-moves (rest (first full-tags))
                                     stack))
                (list-tag-stack-pairs-help (rest full-tags)
                                           stack)))))
                                           
;;returns how many closes you need
(define (last-move stack)
  (/ (- (foldl (lambda (x y) (+ 1 y)) 0 stack) 1) 2))


;;obsolete
(define (make-slashed tags)
  (map (lambda (tag)
         (string->symbol
          (string-append "/" (symbol->string tag))))
       tags))
  

;;stack: bottom: context/tags/context/.../context
;;       the bottom context in a stack is the initial context.
;;slash-list is a list of stuff that we've slashed. how fitting.
;;returns a list of list of tags to get to the goal context


(define (several-moves tag-list stack)
  (foldl make-move stack tag-list))
         

;;returns a new stack
(define (make-move open-tag stack)
  ;(display "stack is ")
  ;(display stack)
  ;(newline)
  (cons (context-open-tag (first stack) open-tag)
        (cons open-tag
              stack)))

(define (pop-move s)
  (rest (rest s)))
                 

;;this is the hack simple version
;; dave tucker should fix this.
;(define (tags-to-diff-context old-c new-c)
;  (cond
;    ((and (not (is-b? old-c)) (is-b? new-c) (not (is-i? old-c)) (is-i? new-c)) '((B I)))
;    ((and (not (is-b? old-c)) (is-b? new-c) (not (is-i? old-c)) (not (is-i? new-c)))
;     '((B)))
;    ((and (not (is-b? old-c)) (not (is-b? new-c)) (not (is-i? old-c)) (not (is-i? new-c)))
;     '(()))
;    ((and (not (is-b? old-c)) (not (is-b? new-c)) (not (is-i? old-c)) (is-i? new-c))
;     '((I)))
;    ((and (not (is-b? old-c)) (not (is-b? new-c)) (is-i? old-c) (is-i? new-c))
;     '(()))
;    ((and (not (is-i? old-c)) (not (is-i? new-c)) (is-b? old-c) (is-b? new-c))
;     '(()))
;    ((and (is-b? old-c) (not (is-b? new-c)) (not (is-i? old-c)) (is-i? new-c)) '((PL I)))
;    ((and (is-b? old-c) (not (is-b? new-c)) (not (is-i? old-c)) (not (is-i? new-c)))
;     '((PL)))
;    ((and (is-b? old-c) (not (is-b? new-c)) (is-i? old-c) (not (is-i? new-c)))
;     '((PL)))
;    ((and (not (is-b? old-c)) (not (is-b? new-c)) (is-i? old-c) (not (is-i? new-c)))
;     '((PL)))
;    ((and (not (is-b? old-c)) (is-b? new-c) (is-i? old-c) (not (is-i? new-c)))
;     '((PL B)))
;    ((and (is-b? old-c) (is-b? new-c) (is-i? old-c) (is-i? new-c))
;     '(()))
;    ((and (is-b? old-c) (is-b? new-c) (not (is-i? old-c)) (is-i? new-c))
;     '((I)))
;    ((and (is-b? old-c) (not (is-b? new-c)) (is-i? old-c) (is-i? new-c))
;     '((PL I)))
;    (else (error "no clause found"))))


(define (display-stack s)
  (cond
    ((empty? (rest s)) (display-context (first s)))
    (else 
     (display-context (first s))
     (display (second s))
     (newline)
     (display-stack (pop-move s)))))

;;  takes in a list of tags, displays them
;; this should, of course, NOT be used in the final version,
;; it's like a temp thing until I get dr's printer working.
(define (display-tags lot)
  (cond
    ((empty? lot) 'done)
    ((eq? '/ (first lot)) (error "you messed up something with logic"))
    ((empty? (rest lot))
     (begin (display "<")
            (display (first lot))
            (display ">")))
    ((eq? (second lot) '/)
     (begin (display "<")
            (display "/")
            (display (first lot))
            (display ">")))
    (else (begin (display "<")
            (display (first lot))
            (display ">")
            (display-tags (rest lot))))))
           
    

(define (pull-context meaning-vector index)
  (meaning-context (vector-ref meaning-vector index)))
          
(define (print-real-answer meaning-vec tag-vec)
  (print-tree (parse-to-tree meaning-vec tag-vec)))


(define initial-stack (cons initial-context '()))

;(define-values (m1 t1) (call-with-input-file "tests/08" read-vectors))

;(display "hello:")
;(display t1)
;(newline)
;;it seems that we need a final context here:

    
;; this is an example run.
;;(define rob-temp-answer (make-answer-space (vector-length m1)))
;;(full-simple-search! m1 random-select rob-temp-answer)

;;(delta initial-stack (pull-context m1 0))


;;(define x (several-moves '(r TT b) initial-stack))

;;(define my-answer (make-answer-space (vector-length m1)))

;;(full-simple-search! m1 smart-random-select my-answer)

;;(display my-answer)



(define (loop-it)
  (full-simple-search! m1 smart-random-select my-answer)
  (if (and
       (empty? (filter (lambda (x)
                        (eq? x 'TT)) (caddr (vector->list my-answer))))
       (> (first (caddr (vector->list my-answer))) 0))
      (display my-answer)
      (loop-it)))