(unit/sig io^
  (import mzlib:function^ context^)
  
  ; --- Starter Reader ---------------------------------------------------------
  
  ; read-vectors : iport -> (vectorof meaning) (vectorof (cons nat (listof sym)))
  (define (read-vectors in)
    (let loop ([context initial-context]
               [context-stack null]
               ; previous space context, if last char was a non-tt space
               [prev-space-context #f]
               [result null]
               [buckets (list (cons 0 null))])
      (let ([c (read-char in)])
        (cond
          [(eof-object? c)
           (values (list->vector (reverse result)) (list->vector (reverse buckets)))]
          [(eq? #\< c)
           (let ([next (read-char in)])
             (if (eq? next #\/)
                 (begin (lex-close-tag in)
                        (loop (car context-stack)
                              (cdr context-stack)
                              prev-space-context
                              result
                              (cons (if (null? (cdar buckets))
                                        (cons (add1 (caar buckets))
                                              (cdar buckets))
                                        ; close after open with no text in between
                                        (cons (caar buckets)
                                              (cddar buckets)))
                                    (cdr buckets))))
                 (let ([open-tag (lex-open-tag in next)])
                   (loop (context-open-tag context open-tag)
                         (cons context context-stack)
                         prev-space-context
                         result
                         (cons (cons (caar buckets)
                                     (cons open-tag (cdar buckets)))
                               (cdr buckets))))))]
          [else (if (is-tt? context)
                    (loop context
                          context-stack
                          #f
                          (cons (lex-chars c context in) result)
                          (cons (cons 0 null)
                                (cons (cons (caar buckets)
                                            (reverse (cdar buckets)))
                                      (cdr buckets))))
                    (let ([sp-context (space-context context)])
                      (let-values ([(meaning ends-in-space)
                                    (lex-chars-non-tt (equal? prev-space-context
                                                              sp-context)
                                                      c context in)])
                        (loop context
                              context-stack
                              (and ends-in-space sp-context)
                              (cons meaning result)
                              (cons (cons 0 null)
                                    (cons (cons (caar buckets)
                                                (reverse (cdar buckets)))
                                          (cdr buckets)))))))]))))
  
  ; --- Lexer ------------------------------------------------------------------
  
  ; lex-chars-non-tt : bool char context iport -> meaning bool
  ; prev-drop-space is true iff the last non-tag character read was a white space
  ; with the same space context as the currrent space context
  ; returns: the string and the whitespace-ness of the last character read
  (define (lex-chars-non-tt prev-drop-space c context in)
    (let ([c-white-space (char-whitespace? c)])
      (let-values ([(last-char-was-space all-space char-lst)
                    (let loop ([drop-space c-white-space]
                               [all-space c-white-space]
                               [char-lst null])
                      (let ([next (peek-char in)])
                        (cond
                          [(or (eq? next #\<) (eof-object? next))
                           (values drop-space all-space (reverse char-lst))]
                          [(char-whitespace? next)
                           (read-char in)
                           (loop #t all-space
                                 (if drop-space
                                     char-lst
                                     (cons next char-lst)))]
                          [else (loop #f #f (cons (read-char in) char-lst))])))])
        (values ((if all-space
                     make-meaning/space
                     make-meaning)
                 context
                 (list->string (if (and prev-drop-space c-white-space)
                                   char-lst
                                   (cons c char-lst))))
                last-char-was-space))))
  
  ; lex-chars : char context iport -> str
  (define (lex-chars c context in)
    (let loop ([all-space (char-whitespace? c)]
	       [result (list c)])
      (let ([next (peek-char in)])
	(if (or (eq? next #\<) (eof-object? next))
	    ((if all-space
		 make-meaning/space
		 make-meaning)
             context
             (list->string (reverse result)))
	    (loop (and all-space (char-whitespace? next))
		  (cons (read-char in) result))))))
  
  ; lex-close-tag : iport -> void
  ; pre: initial </ has already been read
  ; post: closing > has been read
  (define (lex-close-tag in)
    (read-char in) ;; first char of closing tag
    (let ([c (read-char in)])
      (unless (eq? c #\>)
        (read-char in))))
  
  (define lex-table
    (make-vector 200 #f))
  
  ; lex-entry : char sym -> void
  (define (lex-entry char response)
    (vector-set! lex-table (char->integer char) response))
  
  [lex-entry #\B 'B]
  [lex-entry #\E 'EM]
  [lex-entry #\I 'I]
  [lex-entry #\P 'PL]
  [lex-entry #\S 'S]
  [lex-entry #\T 'TT]
  [lex-entry #\U 'U]
  [lex-entry #\0 '|0|]
  [lex-entry #\1 '|1|]
  [lex-entry #\2 '|2|]
  [lex-entry #\3 '|3|]
  [lex-entry #\4 '|4|]
  [lex-entry #\5 '|5|]
  [lex-entry #\6 '|6|]
  [lex-entry #\7 '|7|]
  [lex-entry #\8 '|8|]
  [lex-entry #\9 '|9|]
  [lex-entry #\r 'r]
  [lex-entry #\g 'g]
  [lex-entry #\b 'b]
  [lex-entry #\c 'c]
  [lex-entry #\m 'm]
  [lex-entry #\y 'y]
  [lex-entry #\k 'k]
  [lex-entry #\w 'w]
  
  ; lex-open-tag : iport char -> sym
  ; pre: initial < and first char of tag have been read
  ; post: closing > has been read
  (define (lex-open-tag in c)
    (unless (eq? (read-char in) #\>)
      (read-char in))
    (vector-ref lex-table (char->integer c)))
  
  )
