
(define (color? s)
  (case s
    ((r g b c m y k w) #t)
    (else #f)))

(define (size? s)
  (case s
    ((|0| |1| |2| |3| |4| |5| |6| |7| |8| |9|) #t)
    (else #f)))

(define (category s)
  (cond ((color? s) 'color)
        ((size? s) 'size)
        (else
         (case s
           ((B I TT U) 'normal)
           ((EM PL S) 'abnormal)
           (else 'unknown)))))


(define (last lst)
  (if (null? lst)
      '()
      (if (null? (cdr lst))
          (car lst)
          (last (cdr lst)))))


(define (all-but-last k lst)
  (if (null? lst)
      (k '())
      (if (null? (cdr lst))
          (k '())
          (all-but-last (lambda (a) (k (cons (car lst) a))) (cdr lst)))))

(define (all-but-last1 lst)
	(all-but-last (lambda (x) x) lst))

(define (cost k tree)
  (cond ((string? tree) (k 0))
        ((symbol? tree) (k (+ (* (string-length (symbol->string tree)) 2) 5)))
        ((null? tree) (k 0))
        ((list? tree) (cost (lambda (a) (cost (lambda (b) (k (+ a b)))
                                              (cdr tree)))
                            (car tree)))))

(define (get-cost tree)
  (cost (lambda (x) x) tree))




(define (opt-doubled tree)
  (let opt-doubled ([k (lambda (x) x)] [tree tree])
    (letrec
        ((do-opt
          (lambda (k lst)
            (if (null? lst)
                (k '())
                (if (and (not (null? (cdr lst)))
                         (list? (car lst))
                         (list? (cadr lst))
                         (eq? (caar lst) (caadr lst)))
                    (do-opt (lambda (a)
                              (do-opt (lambda (b)
                                        (do-opt (lambda (c)
                                                  (k (cons (cons (caar lst)
                                                                 (append a b))
                                                           c)))
                                                (cddr lst)))
                                      (cdadr lst)))
                            (cdar lst))
                    (opt-doubled (lambda (a)
                                   (do-opt (lambda (b) (k (cons a b)))
                                           (cdr lst)))
                                 (car lst)))))))
      (cond ((string? tree) (k tree))
            ((list? tree)
             (do-opt (lambda (a) (k (cons (car tree) a)))
                     (cdr tree)))))))



(define (opt-sandwitch tree)
  (let opt-sandwitch ([k (lambda (x) x)] [tree tree])
    (letrec
        ((checkpos
          (lambda (stuff)
            (and (list? stuff)
                 (not (null? stuff))
                 (list? (car stuff))
                 (or (color? (caar stuff))
                     (size? (caar stuff))))))
         (do-opt
          (lambda (k lst)
            (if (null? lst)
                (k '())
                (if (and (list? lst)
                         (checkpos lst)
                         (checkpos (cdr lst))
                         (checkpos (cddr lst))
                         (eq? (category (caar lst)) (category (caadr lst)))
                         (eq? (category (caar lst)) (category (caaddr lst)))
                         (eq? (caar lst) (caaddr lst)))
                    (do-opt (lambda (a)
                              (opt-sandwitch (lambda (b)
                                               (do-opt (lambda (c)
                                                         (do-opt (lambda (d)
                                                                   (k (cons (cons (caar lst)
                                                                                  (append a (cons b c)))
                                                                            d)))
                                                                 (cdddr lst)))
                                                       (cdaddr lst)))
                                             (cadr lst)))
                            (cdar lst))
                    (opt-sandwitch (lambda (a)
                                     (do-opt (lambda (b) (k (cons a b)))
                                             (cdr lst)))
                                   (car lst)))))))
      (cond ((string? tree) (k tree))
            ((list? tree)
             (do-opt (lambda (a) (k (cons (car tree) a)))
                     (cdr tree)))))))



(define (opt-transpose-right tree agg)
  (let opt-transpose-right ([k (lambda (x) x)] [tree tree] [agg agg])
    (letrec
        ((do-opt
          (lambda (k lst)
            (cond ((null? lst) (k '()))
                  ((null? (cdr lst))
                   (opt-transpose-right (lambda (a) (k (list a)))
                                        (car lst) agg))
                  (else
                   (let ((f (car lst))
                         (s (cadr lst)))
                     (if (and (list? f)
                              (list? s)
                              (list? (cadr s))
                              (eq? (car f) (caadr s))
                              (eq? (category (car f)) 'normal)
                              (eq? (category (car s)) 'normal)
                              (or agg
                                  (null? (cddr s)) ;; don't double B, so always save
                                  (>= (cost (lambda (x) x)
                                            (car f))
                                      (cost (lambda (x) x)
                                            (car s)))))
                         (do-opt (lambda (rest)
                                   (do-opt (lambda (a)
                                             (do-opt (lambda (b)
                                                       (do-opt (lambda (c)
                                                                 (let ((chunk1
                                                                        (cons (car f)
                                                                              (append a (list (cons (car s) b)))))
                                                                       (chunk2
                                                                        (cons (car s) c)))
                                                                   (if (null? (cddr s))
                                                                       (k (cons chunk1 rest))
                                                                       (k (cons chunk1 (cons chunk2 rest))))))
                                                               (cddr s)))
                                                     (cdadr s)))
                                           (cdr f)))
                                 (cddr lst))
                         (opt-transpose-right (lambda (a)
                                                (do-opt (lambda (b) (k (cons a b)))
                                                        (cdr lst)))
                                              (car lst) agg))))))))
      (cond ((string? tree) (k tree))
            ((list? tree)
             (do-opt (lambda (a) (k (cons (car tree) a)))
                     (cdr tree)))))))




(define (opt-transpose-left tree agg)
  (let opt-transpose-left ([k (lambda (x) x)] [tree tree] [agg agg])
    (letrec
        ((do-opt
          (lambda (k lst)
            (cond ((null? lst) (k '()))
                  ((null? (cdr lst))
                   (opt-transpose-left (lambda (a) (k (list a)))
                                       (car lst) agg))
                  (else
                   (let* ((f (car lst))
                          (s (cadr lst))
                          (lf (if (list? f) (last f) #f)))
                     ((lambda (k2)
                        (if (list? f) (all-but-last k2 (cdr f)) (k2 #f)))
                      (lambda (x)
                        (if (and (list? f)
                                 (list? s)
                                 (list? lf)
                                 (eq? (car lf) (car s))
                                 (eq? (category (car f)) 'normal)
                                 (eq? (category (car s)) 'normal)
                                 (or agg
                                     (null? x) ;; x empty, so always save
                                     (cost (lambda (a)
                                             (cost (lambda (b) (<= a b))
                                                   (car s)))
                                           (car f))))
                            (do-opt (lambda (rest)
                                      (do-opt (lambda (c)
                                                (do-opt (lambda (dee)
                                                          (do-opt (lambda (eee)
                                                                    (let ((chunk1
                                                                           (cons (car s)
                                                                                 (cons (cons (car f) c)
                                                                                       dee)))
                                                                          (chunk2
                                                                           (cons (car f)
                                                                                 eee)))
                                                                      (k (if (null? x)
                                                                             (cons chunk1 rest)
                                                                             (cons chunk2 (cons chunk1 rest))))))
                                                                  x))
                                                        (cdr s)))
                                              (cdr lf)))
                                    (cddr lst))
                            (opt-transpose-left (lambda (a)
                                                  (do-opt (lambda (b) (k (cons a b)))
                                                          (cdr lst)))
                                                (car lst) agg))))))))))
      (cond ((string? tree) (k tree))
            ((list? tree) (do-opt (lambda (a) (k (cons (car tree) a)))
                                  (cdr tree)))))))



(define (apply-single-pass tree)
  (letrec
      ((apply-opts
        (lambda (tree opts)
          (if (null? opts)
              tree
              (begin (collect-garbage)
                     (let ((nt ((car opts) tree)))
                       (update-tree nt)
                       (apply-opts nt (cdr opts))))))))
    (apply-opts tree *opt-list*)))


(define (run-all-opts tree)
  (run-all-opts (apply-single-pass tree)))


(define *opt-list*
  (list
   opt-doubled
   (lambda (tree) (opt-transpose-right tree #f))
   opt-doubled
   (lambda (tree) (opt-transpose-left tree #f))
   opt-sandwitch
   ))

