(unit/sig timer^
  (import io^ search^ incremental^ opts^ (argv))
  
  ; start-msec must be computed before anything else
  (define start-msec (current-milliseconds))
  (define total-msec (* 1000 (- (or (string->number
                                  (or (getenv "TLIMIT")
                                      (and (> (vector-length argv) 0)
                                           (vector-ref argv 0))
                                      "180"))
                                 180) 5)))
  (define stop-msec (+ start-msec total-msec))
  
  ; These are fairly arbitrary values.  They must leave time for these tasks:
  ; - possibly reading the original tree
  ; - tree-optimizations
  ; - printing the answer
  (define WRITE-MSEC-LEFT 30000)
  
  (define-values (meaning-vector original-tag-vector)
    (read-vectors (current-input-port)))
  
  (define avail-msec (- total-msec (- (current-milliseconds) start-msec)))

  (define KILL-MSEC-LEFT (+ WRITE-MSEC-LEFT (* avail-msec 0.25)))

  (collect-garbage)
  
  ; more here - original cost is too high, since we eliminated white space
  ; and immediately closed tags
  (define tags-cost-pair (cons original-tag-vector
                               (rob-cost original-tag-vector
				         (vector-length original-tag-vector))))
  (define tags-semaphore (make-semaphore 1))
  
  (collect-garbage)
  
  (define best-tree #f)
  (define tree-semaphore (make-semaphore 1))
  
  ; update-if-better : (vectorof (cons nat (listof sym))) nat -> void
  (define (update-if-better v cost)
    (semaphore-wait tags-semaphore)
    (when (< cost (cdr tags-cost-pair))
      (set! tags-cost-pair (cons (copy-vector v) cost)))
    (semaphore-post tags-semaphore))
  
  (define (update-tree tree)
    (semaphore-wait tree-semaphore)
    (set! best-tree tree)
    (semaphore-post tree-semaphore))
  
  ; copy-vector : vector -> vector
  (define (copy-vector v)
    (let* ([len (vector-length v)]
           [new-v (make-vector len (vector-ref v 0))])
      (let loop ([i (sub1 len)])
        (unless (zero? i)
          (vector-set! new-v i (vector-ref v i))
          (loop (sub1 i))))
      new-v))
  
  (define done (make-semaphore 0))
  
  (define (run)
    (let ([cust (make-custodian)])
      (parameterize ([current-custodian cust])
	(thread (lambda ()
		  (let kill-loop ()
		    (cond
		      [(< (- stop-msec (current-milliseconds)) KILL-MSEC-LEFT)
		       (semaphore-post done)
		       (custodian-shutdown-all cust)]
		      [else (sleep 1) (kill-loop)]))))
	(thread (lambda ()
		  (search meaning-vector)))))
    
    (semaphore-wait done)
    
    ; Do we compute the final tree cost first and compare to original cost?
    ; That would take time, but reduce the chance of messing up.
    
    (set! best-tree (parse-to-tree meaning-vector (car tags-cost-pair)))
    
    (let ([cust (make-custodian)])
      (parameterize ([current-custodian cust])
	(thread (lambda ()
		  (let kill-loop ()
		    (cond
		      [(< (- stop-msec (current-milliseconds)) WRITE-MSEC-LEFT)
		       (semaphore-post done)
		       (custodian-shutdown-all cust)]
		      [else (sleep 1) (kill-loop)]))))
	(thread (lambda ()
                  (run-all-opts best-tree)
                  ))))
    
    (semaphore-wait done)
    
    ; more here - call tree optimiser
    
    (print-tree best-tree)
    
    )
  )